# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import TYPE_CHECKING
from logging import Logger
from abc import ABC

if TYPE_CHECKING:
    from .client import MaubotMatrixClient
    from .command_spec import CommandSpec


class Plugin(ABC):
    client: 'MaubotMatrixClient'
    id: str
    log: Logger

    def __init__(self, client: 'MaubotMatrixClient', plugin_instance_id: str, log: Logger) -> None:
        self.client = client
        self.id = plugin_instance_id
        self.log = log

    def set_command_spec(self, spec: 'CommandSpec') -> None:
        self.client.set_command_spec(self.id, spec)

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Type
from zipfile import ZipFile, BadZipFile
import configparser
import logging
import sys
import os

from ..lib.zipimport import zipimporter, ZipImportError
from ..plugin_base import Plugin
from .abc import PluginLoader, PluginClass, IDConflictError


class MaubotZipImportError(Exception):
    pass


class ZippedPluginLoader(PluginLoader):
    path_cache: Dict[str, 'ZippedPluginLoader'] = {}
    log = logging.getLogger("maubot.loader.zip")

    path: str
    id: str
    version: str
    modules: List[str]
    main_class: str
    main_module: str
    _loaded: Type[PluginClass]
    _importer: zipimporter

    def __init__(self, path: str) -> None:
        super().__init__()
        self.path = path
        self.id = None
        self._loaded = None
        self._importer = None
        self._load_meta()
        self._run_preload_checks(self._get_importer())
        try:
            existing = self.id_cache[self.id]
            raise IDConflictError(f"Plugin with id {self.id} already loaded from {existing.source}")
        except KeyError:
            pass
        self.path_cache[self.path] = self
        self.id_cache[self.id] = self
        self.log.debug(f"Preloaded plugin {self.id} from {self.path}")

    @classmethod
    def get(cls, path: str) -> 'ZippedPluginLoader':
        try:
            return cls.path_cache[path]
        except KeyError:
            return cls(path)

    @property
    def source(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return ("<ZippedPlugin "
                f"path='{self.path}' "
                f"id='{self.id}' "
                f"loaded={self._loaded is not None}>")

    def _load_meta(self) -> None:
        try:
            file = ZipFile(self.path)
            data = file.read("maubot.ini")
        except FileNotFoundError as e:
            raise MaubotZipImportError("Maubot plugin not found") from e
        except BadZipFile as e:
            raise MaubotZipImportError("File is not a maubot plugin") from e
        except KeyError as e:
            raise MaubotZipImportError("File does not contain a maubot plugin definition") from e
        config = configparser.ConfigParser()
        try:
            config.read_string(data.decode("utf-8"), source=f"{self.path}/maubot.ini")
            meta = config["maubot"]
            meta_id = meta["ID"]
            version = meta["Version"]
            modules = [mod.strip() for mod in meta["Modules"].split(",")]
            main_class = meta["MainClass"]
            main_module = modules[-1]
            if "/" in main_class:
                main_module, main_class = main_class.split("/")[:2]
        except (configparser.Error, KeyError, IndexError, ValueError) as e:
            raise MaubotZipImportError("Maubot plugin definition in file is invalid") from e
        if self.id and meta_id != self.id:
            raise MaubotZipImportError("Maubot plugin ID changed during reload")
        self.id, self.version, self.modules = meta_id, version, modules
        self.main_class, self.main_module = main_class, main_module

    def _get_importer(self, reset_cache: bool = False) -> zipimporter:
        try:
            if not self._importer:
                self._importer = zipimporter(self.path)
            if reset_cache:
                self._importer.reset_cache()
            return self._importer
        except ZipImportError as e:
            raise MaubotZipImportError("File not found or not a maubot plugin") from e

    def _run_preload_checks(self, importer: zipimporter) -> None:
        try:
            code = importer.get_code(self.main_module.replace(".", "/"))
            if self.main_class not in code.co_names:
                raise MaubotZipImportError(
                    f"Main class {self.main_class} not in {self.main_module}")
        except ZipImportError as e:
            raise MaubotZipImportError(
                f"Main module {self.main_module} not found in file") from e
        for module in self.modules:
            try:
                importer.find_module(module)
            except ZipImportError as e:
                raise MaubotZipImportError(f"Module {module} not found in file") from e

    def load(self, reset_cache: bool = False) -> Type[PluginClass]:
        if self._loaded is not None and not reset_cache:
            return self._loaded
        importer = self._get_importer(reset_cache=reset_cache)
        self._run_preload_checks(importer)
        if reset_cache:
            self.log.debug(f"Preloaded plugin {self.id} from {self.path}")
        for module in self.modules:
            importer.load_module(module)
        main_mod = sys.modules[self.main_module]
        plugin = getattr(main_mod, self.main_class)
        if not issubclass(plugin, Plugin):
            raise MaubotZipImportError("Main class of plugin does not extend maubot.Plugin")
        self._loaded = plugin
        self.log.debug(f"Loaded and imported plugin {self.id} from {self.path}")
        return plugin

    def reload(self) -> Type[PluginClass]:
        self.unload()
        return self.load(reset_cache=True)

    def unload(self) -> None:
        for name, mod in list(sys.modules.items()):
            if getattr(mod, "__file__", "").startswith(self.path):
                del sys.modules[name]
        self._loaded = None
        self.log.debug(f"Unloaded plugin {self.id} at {self.path}")

    def destroy(self) -> None:
        self.unload()
        try:
            del self.path_cache[self.path]
        except KeyError:
            pass
        try:
            del self.id_cache[self.id]
        except KeyError:
            pass
        self.id = None
        self.path = None
        self.version = None
        self.modules = None
        if self._importer:
            self._importer.remove_cache()
            self._importer = None
        self._loaded = None

    @classmethod
    def load_all(cls, *args: str) -> None:
        cls.log.debug("Preloading plugins...")
        for directory in args:
            for file in os.listdir(directory):
                if not file.endswith(".mbp"):
                    continue
                path = os.path.join(directory, file)
                try:
                    ZippedPluginLoader.get(path)
                except (MaubotZipImportError, IDConflictError):
                    cls.log.exception(f"Failed to load plugin at {path}")



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Type
from sqlalchemy import (Column, String, Boolean, ForeignKey, Text, TypeDecorator)
from sqlalchemy.orm import Query, scoped_session
from sqlalchemy.ext.declarative import declarative_base
import json

from mautrix.types import UserID, FilterID, SyncToken, ContentURI
from mautrix.client.api.types.util import Serializable
from mautrix import ClientStore

from .command_spec import CommandSpec

Base: declarative_base = declarative_base()


def make_serializable_alchemy(serializable_type: Type[Serializable]):
    class SerializableAlchemy(TypeDecorator):
        impl = Text

        @property
        def python_type(self):
            return serializable_type

        def process_literal_param(self, value: Serializable, _) -> str:
            return json.dumps(value.serialize()) if value is not None else None

        def process_bind_param(self, value: Serializable, _) -> str:
            return json.dumps(value.serialize()) if value is not None else None

        def process_result_value(self, value: str, _) -> serializable_type:
            return serializable_type.deserialize(json.loads(value)) if value is not None else None

    return SerializableAlchemy


class DBPlugin(Base):
    query: Query
    __tablename__ = "plugin"

    id: str = Column(String(255), primary_key=True)
    type: str = Column(String(255), nullable=False)
    enabled: bool = Column(Boolean, nullable=False, default=False)
    primary_user: UserID = Column(String(255),
                                  ForeignKey("client.id", onupdate="CASCADE", ondelete="RESTRICT"),
                                  nullable=False)


class DBClient(Base):
    query: Query
    __tablename__ = "client"

    id: UserID = Column(String(255), primary_key=True)
    homeserver: str = Column(String(255), nullable=False)
    access_token: str = Column(String(255), nullable=False)

    next_batch: SyncToken = Column(String(255), nullable=False, default="")
    filter_id: FilterID = Column(String(255), nullable=False, default="")

    sync: bool = Column(Boolean, nullable=False, default=True)
    autojoin: bool = Column(Boolean, nullable=False, default=True)

    displayname: str = Column(String(255), nullable=False, default="")
    avatar_url: ContentURI = Column(String(255), nullable=False, default="")


class DBCommandSpec(Base):
    query: Query
    __tablename__ = "command_spec"

    plugin: str = Column(String(255),
                         ForeignKey("plugin.id", onupdate="CASCADE", ondelete="CASCADE"),
                         primary_key=True)
    client: UserID = Column(String(255),
                            ForeignKey("client.id", onupdate="CASCADE", ondelete="CASCADE"),
                            primary_key=True)
    spec: CommandSpec = Column(make_serializable_alchemy(CommandSpec), nullable=False)


def init(session: scoped_session) -> None:
    DBPlugin.query = session.query_property()
    DBClient.query = session.query_property()
    DBCommandSpec.query = session.query_property()



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Union, Callable

from mautrix import Client as MatrixClient
from mautrix.client import EventHandler
from mautrix.types import EventType, MessageEvent

from .command_spec import ParsedCommand, CommandSpec


class MaubotMatrixClient(MatrixClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.command_handlers: Dict[str, List[EventHandler]] = {}
        self.commands: List[ParsedCommand] = []
        self.command_specs: Dict[str, CommandSpec] = {}

        self.add_event_handler(self._command_event_handler, EventType.ROOM_MESSAGE)

    def set_command_spec(self, plugin_id: str, spec: CommandSpec) -> None:
        self.command_specs[plugin_id] = spec
        self._reparse_command_specs()

    def _reparse_command_specs(self) -> None:
        self.commands = [parsed_command
                         for spec in self.command_specs.values()
                         for parsed_command in spec.parse()]

    def remove_command_spec(self, plugin_id: str) -> None:
        try:
            del self.command_specs[plugin_id]
            self._reparse_command_specs()
        except KeyError:
            pass

    async def _command_event_handler(self, evt: MessageEvent) -> None:
        for command in self.commands:
            if command.match(evt):
                await self._trigger_command(command, evt)
                return

    async def _trigger_command(self, command: ParsedCommand, evt: MessageEvent) -> None:
        for handler in self.command_handlers.get(command.name, []):
            await handler(evt)

    def on(self, var: Union[EventHandler, EventType, str]
           ) -> Union[EventHandler, Callable[[EventHandler], EventHandler]]:
        if isinstance(var, str):
            def decorator(func: EventHandler) -> EventHandler:
                self.add_command_handler(var, func)
                return func

            return decorator
        return super().on(var)

    def add_command_handler(self, command: str, handler: EventHandler) -> None:
        self.command_handlers.setdefault(command, []).append(handler)

    def remove_command_handler(self, command: str, handler: EventHandler) -> None:
        try:
            self.command_handlers[command].remove(handler)
        except (KeyError, ValueError):
            pass



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from aiohttp import web
import asyncio

from mautrix.api import PathBuilder, Method

from .config import Config
from .__meta__ import __version__


class MaubotServer:
    def __init__(self, config: Config, loop: asyncio.AbstractEventLoop):
        self.loop = loop or asyncio.get_event_loop()
        self.app = web.Application(loop=self.loop)
        self.config = config

        path = PathBuilder(config["server.base_path"])
        self.add_route(Method.GET, path.version, self.version)

        as_path = PathBuilder(config["server.appservice_base_path"])
        self.add_route(Method.PUT, as_path.transactions, self.handle_transaction)

        self.runner = web.AppRunner(self.app)

    def add_route(self, method: Method, path: PathBuilder, handler) -> None:
        self.app.router.add_route(method.value, str(path), handler)

    async def start(self) -> None:
        await self.runner.setup()
        site = web.TCPSite(self.runner, self.config["server.hostname"], self.config["server.port"])
        await site.start()

    async def stop(self) -> None:
        await self.runner.cleanup()

    @staticmethod
    async def version(_: web.Request) -> web.Response:
        return web.json_response({
            "version": __version__
        })

    async def handle_transaction(self, request: web.Request) -> web.Response:
        return web.Response(status=501)



import setuptools
import os

path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "maubot", "__meta__.py")
__version__ = "UNKNOWN"
with open(path) as f:
    exec(f.read())

setuptools.setup(
    name="maubot",
    version=__version__,
    url="https://github.com/maubot/maubot",

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="A plugin-based Matrix bot system.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",

    packages=setuptools.find_packages(),

    install_requires=[
        "mautrix>=0.4,<0.5",
        "aiohttp>=3.0.1,<4",
        "SQLAlchemy>=1.2.3,<2",
        "alembic>=1.0.0,<2",
        "commonmark>=0.8.1,<1",
        "ruamel.yaml>=0.15.35,<0.16",
        "attrs>=18.2.0,<19",
    ],

    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points="""
        [console_scripts]
        maubot=maubot.__main__:main
    """,
    data_files=[
        (".", ["example-config.yaml"]),
    ],
)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Dict, Pattern, Union, Tuple, Optional, Any
from attr import dataclass
import re

from mautrix.types import MessageEvent, MatchedCommand, MatchedPassiveCommand
from mautrix.client.api.types.util import SerializableAttrs


@dataclass
class Argument(SerializableAttrs['Argument']):
    matches: str
    required: bool = False
    description: str = None


@dataclass
class Command(SerializableAttrs['Command']):
    syntax: str
    arguments: Dict[str, Argument]
    description: str = None


@dataclass
class PassiveCommand(SerializableAttrs['PassiveCommand']):
    name: str
    matches: str
    match_against: str
    match_event: MessageEvent = None


class ParsedCommand:
    name: str
    is_passive: bool
    arguments: List[str]
    starts_with: str
    matches: Pattern
    match_against: str
    match_event: MessageEvent

    def __init__(self, command: Union[PassiveCommand, Command]) -> None:
        if isinstance(command, PassiveCommand):
            self._init_passive(command)
        elif isinstance(command, Command):
            self._init_active(command)
        else:
            raise ValueError("Command parameter must be a Command or a PassiveCommand.")

    def _init_passive(self, command: PassiveCommand) -> None:
        self.name = command.name
        self.is_passive = True
        self.match_against = command.match_against
        self.matches = re.compile(command.matches)
        self.match_event = command.match_event

    def _init_active(self, command: Command) -> None:
        self.name = command.syntax
        self.is_passive = False

        regex_builder = []
        sw_builder = []
        argument_encountered = False

        for word in command.syntax.split(" "):
            arg = command.arguments.get(word, None)
            if arg is not None and len(word) > 0:
                argument_encountered = True
                regex = f"({arg.matches})" if arg.required else f"(?:{arg.matches})?"
                self.arguments.append(word)
                regex_builder.append(regex)
            else:
                if not argument_encountered:
                    sw_builder.append(word)
                regex_builder.append(re.escape(word))
        self.starts_with = "!" + " ".join(sw_builder)
        self.matches = re.compile("^!" + " ".join(regex_builder) + "$")
        self.match_against = "body"

    def match(self, evt: MessageEvent) -> bool:
        return self._match_passive(evt) if self.is_passive else self._match_active(evt)

    @staticmethod
    def _parse_key(key: str) -> Tuple[str, Optional[str]]:
        if '.' not in key:
            return key, None
        key, next_key = key.split('.', 1)
        if len(key) > 0 and key[0] == "[":
            end_index = next_key.index("]")
            key = key[1:] + "." + next_key[:end_index]
            next_key = next_key[end_index + 2:] if len(next_key) > end_index + 1 else None
        return key, next_key

    @classmethod
    def _recursive_get(cls, data: Any, key: str) -> Any:
        if not data:
            return None
        key, next_key = cls._parse_key(key)
        if next_key is not None:
            return cls._recursive_get(data[key], next_key)
        return data[key]

    def _match_passive(self, evt: MessageEvent) -> bool:
        try:
            match_against = self._recursive_get(evt.content, self.match_against)
        except KeyError:
            match_against = None
        match_against = match_against or evt.content.body
        matches = [[match.string[match.start():match.end()]] + list(match.groups())
                   for match in self.matches.finditer(match_against)]
        if not matches:
            return False
        if evt.unsigned.passive_command is None:
            evt.unsigned.passive_command = {}
        evt.unsigned.passive_command[self.name] = MatchedPassiveCommand(captured=matches)
        return True

    def _match_active(self, evt: MessageEvent) -> bool:
        if not evt.content.body.startswith(self.starts_with):
            return False
        match = self.matches.match(evt.content.body)
        if not match:
            return False
        evt.content.command = MatchedCommand(matched=self.name,
                                             arguments=dict(zip(self.arguments, match.groups())))
        return True


@dataclass
class CommandSpec(SerializableAttrs['CommandSpec']):
    commands: List[Command] = []
    passive_commands: List[PassiveCommand] = []

    def __add__(self, other: 'CommandSpec') -> 'CommandSpec':
        return CommandSpec(commands=self.commands + other.commands,
                           passive_commands=self.passive_commands + other.passive_commands)

    def parse(self) -> List[ParsedCommand]:
        return [ParsedCommand(command) for command in self.commands + self.passive_commands]



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import random
import string

from mautrix.util import BaseConfig


class Config(BaseConfig):
    @staticmethod
    def _new_token() -> str:
        return "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(64))

    def update(self):
        base, copy, copy_dict = self._pre_update()
        copy("database")
        copy("plugin_directories")
        copy("server.hostname")
        copy("server.port")
        copy("server.listen")
        copy("server.base_path")
        shared_secret = self["server.shared_secret"]
        if shared_secret is None or shared_secret == "generate":
            base["server.shared_secret"] = self._new_token()
        else:
            base["server.shared_secret"] = shared_secret
        copy("logging")



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from sqlalchemy import orm
import sqlalchemy as sql
import logging.config
import argparse
import asyncio
import copy
import sys

from .config import Config
from .db import Base, init as init_db
from .server import MaubotServer
from .client import Client, init as init_client
from .loader import ZippedPluginLoader
from .plugin import PluginInstance
from .__meta__ import __version__

parser = argparse.ArgumentParser(description="A plugin-based Matrix bot system.",
                                 prog="python -m maubot")
parser.add_argument("-c", "--config", type=str, default="config.yaml",
                    metavar="<path>", help="the path to your config file")
parser.add_argument("-b", "--base-config", type=str, default="example-config.yaml",
                    metavar="<path>", help="the path to the example config "
                                           "(for automatic config updates)")
args = parser.parse_args()

config = Config(args.config, args.base_config)
config.load()
config.update()

logging.config.dictConfig(copy.deepcopy(config["logging"]))
log = logging.getLogger("maubot.init")
log.debug(f"Initializing maubot {__version__}")

db_engine: sql.engine.Engine = sql.create_engine(config["database"])
db_factory = orm.sessionmaker(bind=db_engine)
db_session = orm.scoping.scoped_session(db_factory)
Base.metadata.bind = db_engine
Base.metadata.create_all()

loop = asyncio.get_event_loop()

init_db(db_session)
init_client(loop)
server = MaubotServer(config, loop)
ZippedPluginLoader.load_all(*config["plugin_directories"])
plugins = PluginInstance.all()

for plugin in plugins:
    plugin.load()

try:
    loop.run_until_complete(asyncio.gather(
        server.start(),
        *[plugin.start() for plugin in plugins]))
    log.debug("Startup actions complete, running forever.")
    loop.run_forever()
except KeyboardInterrupt:
    log.debug("Keyboard interrupt received, stopping...")
    for client in Client.cache.values():
        client.stop()
    db_session.commit()
    loop.run_until_complete(server.stop())
    sys.exit(0)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Optional
import logging

from mautrix.types import UserID

from .db import DBPlugin
from .client import Client
from .loader import PluginLoader
from .plugin_base import Plugin

log = logging.getLogger("maubot.plugin")


class PluginInstance:
    cache: Dict[str, 'PluginInstance'] = {}
    plugin_directories: List[str] = []

    log: logging.Logger
    loader: PluginLoader
    client: Client
    plugin: Plugin

    def __init__(self, db_instance: DBPlugin):
        self.db_instance = db_instance
        self.log = logging.getLogger(f"maubot.plugin.{self.id}")
        self.cache[self.id] = self

    def load(self) -> None:
        try:
            self.loader = PluginLoader.find(self.type)
        except KeyError:
            self.log.error(f"Failed to find loader for type {self.type}")
            self.db_instance.enabled = False
            return
        self.client = Client.get(self.primary_user)
        if not self.client:
            self.log.error(f"Failed to get client for user {self.primary_user}")
            self.db_instance.enabled = False

    async def start(self) -> None:
        self.log.debug(f"Starting...")
        cls = self.loader.load()
        self.plugin = cls(self.client.client, self.id, self.log)
        self.loader.references |= {self}
        await self.plugin.start()

    async def stop(self) -> None:
        self.log.debug("Stopping...")
        self.loader.references -= {self}
        await self.plugin.stop()
        self.plugin = None

    @classmethod
    def get(cls, instance_id: str, db_instance: Optional[DBPlugin] = None
            ) -> Optional['PluginInstance']:
        try:
            return cls.cache[instance_id]
        except KeyError:
            db_instance = db_instance or DBPlugin.query.get(instance_id)
            if not db_instance:
                return None
            return PluginInstance(db_instance)

    @classmethod
    def all(cls) -> List['PluginInstance']:
        return [cls.get(plugin.id, plugin) for plugin in DBPlugin.query.all()]

    # region Properties

    @property
    def id(self) -> str:
        return self.db_instance.id

    @id.setter
    def id(self, value: str) -> None:
        self.db_instance.id = value

    @property
    def type(self) -> str:
        return self.db_instance.type

    @type.setter
    def type(self, value: str) -> None:
        self.db_instance.type = value

    @property
    def enabled(self) -> bool:
        return self.db_instance.enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.db_instance.enabled = value

    @property
    def primary_user(self) -> UserID:
        return self.db_instance.primary_user

    @primary_user.setter
    def primary_user(self, value: UserID) -> None:
        self.db_instance.primary_user = value

    # endregion



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Optional
from aiohttp import ClientSession
import asyncio
import logging

from mautrix.types import UserID, SyncToken, FilterID, ContentURI, StateEvent, Membership, EventType

from .db import DBClient
from .matrix import MaubotMatrixClient

log = logging.getLogger("maubot.client")


class Client:
    loop: asyncio.AbstractEventLoop
    cache: Dict[UserID, 'Client'] = {}
    http_client: ClientSession = None

    db_instance: DBClient
    client: MaubotMatrixClient

    def __init__(self, db_instance: DBClient) -> None:
        self.db_instance = db_instance
        self.cache[self.id] = self
        self.log = log.getChild(self.id)
        self.client = MaubotMatrixClient(mxid=self.id, base_url=self.homeserver,
                                         token=self.access_token, client_session=self.http_client,
                                         log=self.log, loop=self.loop, store=self.db_instance)
        if self.autojoin:
            self.client.add_event_handler(self._handle_invite, EventType.ROOM_MEMBER)

    def start(self) -> None:
        asyncio.ensure_future(self._start(), loop=self.loop)

    async def _start(self) -> None:
        try:
            await self.client.start()
        except Exception:
            self.log.exception("Fail")

    def stop(self) -> None:
        self.client.stop()

    @classmethod
    def get(cls, user_id: UserID, db_instance: Optional[DBClient] = None) -> Optional['Client']:
        try:
            return cls.cache[user_id]
        except KeyError:
            db_instance = db_instance or DBClient.query.get(user_id)
            if not db_instance:
                return None
            return Client(db_instance)

    @classmethod
    def all(cls) -> List['Client']:
        return [cls.get(user.id, user) for user in DBClient.query.all()]

    async def _handle_invite(self, evt: StateEvent) -> None:
        if evt.state_key == self.id and evt.content.membership == Membership.INVITE:
            await self.client.join_room_by_id(evt.room_id)

    # region Properties

    @property
    def id(self) -> UserID:
        return self.db_instance.id

    @property
    def homeserver(self) -> str:
        return self.db_instance.homeserver

    @property
    def access_token(self) -> str:
        return self.db_instance.access_token

    @access_token.setter
    def access_token(self, value: str) -> None:
        self.client.api.token = value
        self.db_instance.access_token = value

    @property
    def next_batch(self) -> SyncToken:
        return self.db_instance.next_batch

    @next_batch.setter
    def next_batch(self, value: SyncToken) -> None:
        self.db_instance.next_batch = value

    @property
    def filter_id(self) -> FilterID:
        return self.db_instance.filter_id

    @filter_id.setter
    def filter_id(self, value: FilterID) -> None:
        self.db_instance.filter_id = value

    @property
    def sync(self) -> bool:
        return self.db_instance.sync

    @sync.setter
    def sync(self, value: bool) -> None:
        self.db_instance.sync = value

    @property
    def autojoin(self) -> bool:
        return self.db_instance.autojoin

    @autojoin.setter
    def autojoin(self, value: bool) -> None:
        if value == self.db_instance.autojoin:
            return
        if value:
            self.client.add_event_handler(self._handle_invite, EventType.ROOM_MEMBER)
        else:
            self.client.remove_event_handler(self._handle_invite, EventType.ROOM_MEMBER)
        self.db_instance.autojoin = value

    @property
    def displayname(self) -> str:
        return self.db_instance.displayname

    @displayname.setter
    def displayname(self, value: str) -> None:
        self.db_instance.displayname = value

    @property
    def avatar_url(self) -> ContentURI:
        return self.db_instance.avatar_url

    @avatar_url.setter
    def avatar_url(self, value: ContentURI) -> None:
        self.db_instance.avatar_url = value

    # endregion


def init(loop: asyncio.AbstractEventLoop) -> None:
    Client.http_client = ClientSession(loop=loop)
    Client.loop = loop
    for client in Client.all():
        client.start()



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2018 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import TypeVar, Type, Dict, Set, TYPE_CHECKING
from abc import ABC, abstractmethod

from ..plugin_base import Plugin

if TYPE_CHECKING:
    from ..plugin import PluginInstance

PluginClass = TypeVar("PluginClass", bound=Plugin)


class IDConflictError(Exception):
    pass


class PluginLoader(ABC):
    id_cache: Dict[str, 'PluginLoader'] = {}

    references: Set['PluginInstance']
    id: str
    version: str

    def __init__(self):
        self.references = set()

    @classmethod
    def find(cls, plugin_id: str) -> 'PluginLoader':
        return cls.id_cache[plugin_id]

    @property
    @abstractmethod
    def source(self) -> str:
        pass

    @abstractmethod
    def load(self) -> Type[PluginClass]:
        pass

    @abstractmethod
    def reload(self) -> Type[PluginClass]:
        pass

    @abstractmethod
    def unload(self) -> None:
        pass
